@@ -2,7 +2,6 @@
 # Debian task selector, mark II.
 # Copyright 2004-2011 by Joey Hess <joeyh@debian.org>.
 # Licensed under the GPL, version 2 or higher.
-use 5.014;
 use Locale::gettext;
 use Getopt::Long;
 use warnings;
@@ -96,35 +95,21 @@ sub all_tasks {
 	map { read_task_desc($_) } list_task_descs();
 }
 
-sub _info_avail {
-	my %ret = ();
+# Returns a list of all available packages.
+sub list_avail {
+	my @list;
 	# Might be better to use the perl apt bindings, but they are not
 	# currently in base.
 	open (AVAIL, "apt-cache dumpavail|");
 	local $_;
-	my ($package, $section, $priority);
 	while (<AVAIL>) {
 		chomp;
-		if (not $_) {
-			if (defined $package && defined $priority && defined $section) {
-				$ret{$package} = {
-				       	"priority" => $priority,
-					"section" => $section,
-				};
-			}
-		}
-		elsif (/^Package: (.*)/) {
-			$package = $1;
-		}
-		elsif (/^Priority: (.*)/) {
-			$priority = $1;
-		}
-		elsif (/^Section: (.*)/) {
-			$section = $1;
+		if (/^Package: (.*)/) {
+			push @list, $1;
 		}
 	}
 	close AVAIL;
-	return %ret;
+	return @list;
 }
 
 # Returns a list of all installed packages.
@@ -140,21 +125,18 @@ sub list_installed {
 	return @list;
 }
 
-my %_info_avail_cache;
-
-# Returns a hash of all available packages.
-sub info_avail {
-	my $package = shift;
-	if (!%_info_avail_cache) {
-		%_info_avail_cache = _info_avail();
-	}
-	return \%_info_avail_cache;
-}
-
+my %avail_pkgs;
 # Given a package name, checks to see if it's available. Memoised.
 sub package_avail {
-	my $package = shift;
-	return info_avail()->{$package} || package_installed($package);
+	my $package=shift;
+	
+	if (! %avail_pkgs) {
+		foreach my $pkg (list_avail()) {
+			$avail_pkgs{$pkg} = 1;
+		}
+	}
+
+	return $avail_pkgs{$package} || package_installed($package);
 }
 
 my %installed_pkgs;
@@ -207,8 +189,12 @@ sub task_installed {
 }
 
 # Given task hash, returns a list of all available packages in the task.
+# If the aptitude_tasks parameter is true, then it does not expand tasks
+# that aptitude knows about, and just returns aptitude task syntax for
+# those.
 sub task_packages {
 	my $task=shift;
+	my $aptitude_tasks=shift;
 	
 	my %list;
 
@@ -220,19 +206,34 @@ sub task_packages {
 	if (! defined $task->{packages}) {
 		# only key
 	}
-	elsif ($task->{packages} eq 'standard') {
-		while (my ($package, $info) = each(info_avail())) {
-			my ($priority, $section) = ($info->{priority}, $info->{section});
-			if (($priority eq 'required' ||
-			     $priority eq 'important' ||
-			     $priority eq 'standard') &&
-		            # Exclude packages in non-main and library sections
-		            $section !~ /^lib|\// &&
-			    # Exclude already installed packages
-		            !package_installed($package)) {
-				$list{$package} = 1;
-			}
+	elsif ($task->{packages} eq 'task-fields') {
+		# task-fields method is built-in for speed and to support
+		# aptitude task definitions
+		if ($aptitude_tasks) {
+			return '~t^'.$task->{task}.'$';
 		}
+		else {
+			local $/="\n\n";
+			open (AVAIL, "apt-cache dumpavail|");
+			while (<AVAIL>) {
+				if (/^Task: (.*)/m) {
+					my @tasks=split(", ", $1);
+					if (grep { $_ eq $task->{task} } @tasks) { 
+						$list{$1}=1 if /^Package: (.*)/m;
+					}
+				}
+			}
+			close AVAIL;
+		}
+	}
+	elsif ($task->{packages} eq 'standard') {
+		# standard method is built in since it cannot easily be
+		# implemented externally.
+		return "~pstandard", "~prequired", "~pimportant";
+	}
+	elsif ($task->{packages} eq 'manual') {
+		# manual package selection is a special case
+		return;
 	}
 	else {
 		# external method
@@ -393,8 +394,8 @@ sub task_script {
 
 sub usage {
 	print STDERR gettext(q{Usage:
-tasksel install <task>...
-tasksel remove <task>...
+tasksel install <task>
+tasksel remove <task>
 tasksel [options]
 	-t, --test          test mode; don't really do anything
 	    --new-install   automatically install some tasks
@@ -415,146 +416,22 @@ sub getopts {
 		exit(1);
 	}
 	# Special case apt-like syntax.
+	if (@ARGV && $ARGV[0] eq "install") {
+		shift @ARGV;
+		$ret{install} = shift @ARGV;
+	}
+	if (@ARGV && $ARGV[0] eq "remove") {
+		shift @ARGV;
+		$ret{remove} = shift @ARGV;
+	}
 	if (@ARGV) {
-		my $cmd = shift @ARGV;
-		if ($cmd eq "install") {
-			$ret{cmd_install} = \@ARGV;
-		}
-		elsif ($cmd eq "remove") {
-			$ret{cmd_remove} = \@ARGV;
-		}
-		else {
-			usage();
-			exit 1;
-		}
+		usage();
+		exit 1;
 	}
 	$testmode=1 if $ret{test}; # set global
 	return %ret;
 }
 
-sub interactive {
-	my $options = shift;
-	my @tasks = @_;
-
-	if (! $options->{"new-install"}) {
-		# Don't install hidden tasks if this is not a new install.
-		map { $_->{_install} = 0 } grep { $_->{_display} == 0 } @tasks;
-	}
-
-	my @list = order_for_display(grep { $_->{_display} == 1 } @tasks);
-	if (@list) {
-		if (! $options->{"new-install"}) {
-			# Find tasks that are already installed.
-			map { $_->{_installed} = task_installed($_) } @list;
-			# Don't install new tasks unless manually selected.
-			map { $_->{_install} = 0 } @list;
-		}
-		else {
-			# Assume that no tasks are installed, to ensure
-			# that complete tasks get installed on new
-			# installs.
-			map { $_->{_installed} = 0 } @list;
-		}
-		my $question="tasksel/tasks";
-		if ($options->{"new-install"}) {
-			$question="tasksel/first";
-		}
-		my @default = grep { $_->{_display} == 1 && ($_->{_install} == 1 || $_->{_installed} == 1) } @tasks;
-		my $tmpfile=`tempfile`;
-		chomp $tmpfile;
-		my $ret=system($debconf_helper, $tmpfile,
-			task_to_debconf_C(@list),
-			task_to_debconf(@list),
-			task_to_debconf_C(@default),
-			$question) >> 8;
-		if ($ret == 30) {
-			exit 10; # back up
-		}
-		elsif ($ret != 0) {
-			error "debconf failed to run";
-		}
-		open(IN, "<$tmpfile");
-		$ret=<IN>;
-		if (! defined $ret) {
-			die "tasksel canceled\n";
-		}
-		chomp $ret;
-		close IN;
-		unlink $tmpfile;
-		
-		# Set _install flags based on user selection.
-		map { $_->{_install} = 0 } @list;
-		foreach my $task (list_to_tasks($ret, @tasks)) {
-			if (! $task->{_installed}) {
-				$task->{_install} = 1;
-			}
-			$task->{_selected} = 1;
-		}
-		foreach my $task (@list) {
-			if (! $task->{_selected} && $task->{_installed}) {
-				$task->{_remove} = 1;
-			}
-		}
-	}
-
-	# If an enhancing task is already marked for
-	# install, probably by preseeding, mark the tasks
-	# it enhances for install.
-	foreach my $task (grep { $_->{_install} && exists $_->{enhances} &&
-	                         length $_->{enhances} } @tasks) {
-		map { $_->{_install}=1 } list_to_tasks($task->{enhances}, @tasks);
-	}
-
-	# Select enhancing tasks for install.
-	# XXX FIXME ugly hack -- loop until enhances settle to handle
-	# chained enhances. This is ugly and could loop forever if
-	# there's a cycle.
-	my $enhances_needswork=1;
-	my %tested;
-	while ($enhances_needswork) {
-		$enhances_needswork=0;
-		foreach my $task (grep { ! $_->{_install} && exists $_->{enhances} &&
-		                         length $_->{enhances} } @tasks) {
-			my %tasknames = map { $_->{task} => $_ } @tasks;
-			my @deps=map { $tasknames{$_} } split ", ", $task->{enhances};
-
-			if (grep { ! defined $_ } @deps) {
-				# task enhances an unavailable or
-				# uninstallable task
-				next;
-			}
-
-			if (@deps) {
-				my $orig_state=$task->{_install};
-
-				# Mark enhancing tasks for install if their
-				# dependencies are met and their test fields
-				# mark them for install.
-				if (! exists $tested{$task->{task}}) {
-					$ENV{TESTING_ENHANCER}=1;
-					task_test($task, $options->{"new-install"}, 0, 1);
-					delete $ENV{TESTING_ENHANCER};
-					$tested{$task->{task}}=$task->{_install};
-				}
-				else {
-					$task->{_install}=$tested{$task->{task}};
-				}
-
-				foreach my $dep (@deps) {
-					if (! $dep->{_install}) {
-						$task->{_install} = 0;
-					}
-				}
-
-				if ($task->{_install} != $orig_state) {
-					$enhances_needswork=1;
-				}
-			}
-		}
-	}
-
-}
-
 sub main {
 	my %options=getopts();
 	my @tasks_remove;
@@ -597,43 +474,188 @@ sub main {
 		exit(0);
 	}
 	
-	if ($options{cmd_install}) {
-		@tasks_install = map { name_to_task($_, @tasks) } @{$options{cmd_install}};
+	if (! $options{"new-install"}) {
+		# Don't install hidden tasks if this is not a new install.
+		map { $_->{_install} = 0 } grep { $_->{_display} == 0 } @tasks;
 	}
-	elsif ($options{cmd_remove}) {
-		@tasks_remove = map { name_to_task($_, @tasks) } @{$options{cmd_remove}};
+	if ($options{"install"}) {
+		my $task=name_to_task($options{"install"}, @tasks);
+		$task->{_install} = 1 if $task;
+	}
+	if ($options{"remove"}) {
+		my $task=name_to_task($options{"remove"}, @tasks);
+		push @tasks_remove, $task;
+	}
+	
+	# The interactive bit.
+	my $interactive=0;
+	my @list = order_for_display(grep { $_->{_display} == 1 } @tasks);
+	if (@list && ! $options{install} && ! $options{remove}) {
+		$interactive=1;
+		if (! $options{"new-install"}) {
+			# Find tasks that are already installed.
+			map { $_->{_installed} = task_installed($_) } @list;
+			# Don't install new tasks unless manually selected.
+			map { $_->{_install} = 0 } @list;
+		}
+		else {
+			# Assume that no tasks are installed, to ensure
+			# that complete tasks get installed on new
+			# installs.
+			map { $_->{_installed} = 0 } @list;
+		}
+		my $question="tasksel/tasks";
+		if ($options{"new-install"}) {
+			$question="tasksel/first";
+		}
+		my @default = grep { $_->{_display} == 1 && ($_->{_install} == 1 || $_->{_installed} == 1) } @tasks;
+		my $tmpfile=`tempfile`;
+		chomp $tmpfile;
+		my $ret=system($debconf_helper, $tmpfile,
+			task_to_debconf_C(@list),
+			task_to_debconf(@list),
+			task_to_debconf_C(@default),
+			$question) >> 8;
+		if ($ret == 30) {
+			exit 10; # back up
+		}
+		elsif ($ret != 0) {
+			error "debconf failed to run";
+		}
+		open(IN, "<$tmpfile");
+		$ret=<IN>;
+		if (! defined $ret) {
+			die "tasksel canceled\n";
+		}
+		chomp $ret;
+		close IN;
+		unlink $tmpfile;
+		
+		# Set _install flags based on user selection.
+		map { $_->{_install} = 0 } @list;
+		foreach my $task (list_to_tasks($ret, @tasks)) {
+			if (! $task->{_installed}) {
+				$task->{_install} = 1;
+			}
+			$task->{_selected} = 1;
+		}
+		foreach my $task (@list) {
+			if (! $task->{_selected} && $task->{_installed}) {
+				push @tasks_remove, $task;
+			}
+		}
+	}
+
+	# If an enhancing task is already marked for
+	# install, probably by preseeding, mark the tasks
+	# it enhances for install.
+	foreach my $task (grep { $_->{_install} && exists $_->{enhances} &&
+	                         length $_->{enhances} } @tasks) {
+		map { $_->{_install}=1 } list_to_tasks($task->{enhances}, @tasks);
+	}
+
+	# Select enhancing tasks for install.
+	# XXX FIXME ugly hack -- loop until enhances settle to handle
+	# chained enhances. This is ugly and could loop forever if
+	# there's a cycle.
+	my $enhances_needswork=1;
+	my %tested;
+	while ($enhances_needswork) {
+		$enhances_needswork=0;
+		foreach my $task (grep { ! $_->{_install} && exists $_->{enhances} &&
+		                         length $_->{enhances} } @tasks) {
+			my %tasknames = map { $_->{task} => $_ } @tasks;
+			my @deps=map { $tasknames{$_} } split ", ", $task->{enhances};
+
+			if (grep { ! defined $_ } @deps) {
+				# task enhances an unavailable or
+				# uninstallable task
+				next;
+			}
+
+			if (@deps) {
+				my $orig_state=$task->{_install};
+
+				# Mark enhancing tasks for install if their
+				# dependencies are met and their test fields
+				# mark them for install.
+				if (! exists $tested{$task->{task}}) {
+					$ENV{TESTING_ENHANCER}=1;
+					task_test($task, $options{"new-install"}, 0, 1);
+					delete $ENV{TESTING_ENHANCER};
+					$tested{$task->{task}}=$task->{_install};
+				}
+				else {
+					$task->{_install}=$tested{$task->{task}};
+				}
+
+				foreach my $dep (@deps) {
+					if (! $dep->{_install}) {
+						$task->{_install} = 0;
+					}
+				}
+
+				if ($task->{_install} != $orig_state) {
+					$enhances_needswork=1;
+				}
+			}
+		}
+	}
+
+	# Add tasks to install and see if any selected task requires manual
+	# selection.
+	my $manual_selection=0;
+	foreach my $task (grep { $_->{_install} } @tasks) {
+		push @tasks_install, $task;
+		if (defined $task->{packages} && $task->{packages} eq 'manual') {
+			$manual_selection=1;
+		}
+	}
+	
+	my @aptitude;
+	if ($manual_selection) {
+		# Manaul selection and task installs, as best
+		# aptitude can do it currently. Disables use of
+		# debconf-apt-progress.
+		@aptitude="aptitude";
+	}
+	elsif (-x "/usr/bin/debconf-apt-progress") {
+		@aptitude="debconf-apt-progress";
+		push @aptitude, split(' ', $options{'debconf-apt-progress'})
+			if exists $options{'debconf-apt-progress'};
+		push @aptitude, qw{-- aptitude -q};
 	}
 	else {
-		interactive(\%options, @tasks);
-
-		# Add tasks to install
-		@tasks_install = grep { $_->{_install} } @tasks;
-		# Add tasks to remove
-		@tasks_remove = grep { $_->{_remove} } @tasks;
+		@aptitude="aptitude";
 	}
-
-	my @cmd;
-	if (-x "/usr/bin/debconf-apt-progress") {
-		@cmd = "debconf-apt-progress";
-		push @cmd, split(' ', $options{'debconf-apt-progress'})
-			if exists $options{'debconf-apt-progress'};
-		push @cmd, "--";
-	}
-	push @cmd, qw{apt-get -q -y -o APT::Install-Recommends=true -o APT::Get::AutomaticRemove=true install};
-
+	
 	# And finally, act on selected tasks.
-	if (@tasks_install || @tasks_remove) {
+	if (@tasks_install || @tasks_remove || $manual_selection) {
+		my @args;
 		foreach my $task (@tasks_remove) {
-			push @cmd, map { "$_-" } task_packages($task);
+			push @args, map { "$_-" } task_packages($task, 0);
 			task_script($task->{task}, "prerm");
 		}
 		foreach my $task (@tasks_install) {
-			push @cmd, task_packages($task);
+			push @args, task_packages($task, 1);
 			task_script($task->{task}, "preinst");
 		}
-		my $ret=run(@cmd);
-		if ($ret != 0) {
-			error gettext("apt-get failed")." ($ret)";
+		# If the user selected no other tasks and manual package
+		# selection, run aptitude w/o the --visual-preview parameter.
+		if (! @args && $manual_selection) {
+			my $ret=run("aptitude");
+			if ($ret != 0) {
+				error gettext("aptitude failed")." ($ret)";
+			}
+		}
+		else {
+			if ($manual_selection) {
+				unshift @args, "--visual-preview";
+			}
+			my $ret=run(@aptitude, "-y", "install", @args);
+			if ($ret != 0) {
+				error gettext("aptitude failed")." ($ret)";
+			}
 		}
 		foreach my $task (@tasks_remove) {
 			task_script($task->{task}, "postrm");
